#include <stdio.h>
#include <curses.h>
#include <string.h>
#include <stdlib.h>
#include <limits.h>
#include <time.h>
#include <errno.h>
#include <sys/select.h>

static WINDOW *mainwindow;
static WINDOW *screen;
int i = 0;

#define RSSI_NOTHING	99	/* Coming from the modem */
int crssi = -113;		/* Calculated RSSI */
int rrssi = RSSI_NOTHING;	/* Real RSSI */
#define ARSSI	300		/* 300 ticks */
int arssi[ARSSI];

#define ARX	30
#define ATX	30
long arx[ARX], atx[ATX];

long uptime = 0, lastuptime = 0;	/* Seconds */
long rxnow = 0, txnow = 0, rxtotal = 0, txtotal = 0;
char mode[100] = "Unknown";
int srvst = -1;

#ifndef min
#define	min(a, b) (((a) < (b)) ? (a) : (b))
#endif

void
bytestosomething(long b, char *s)
{

	if (b > 20 * 1024 * 1024) {
		sprintf(s, "%2.02f MiB", (double)b / (1024 * 1024));
	} else if (b > 20 * 1024) {
		sprintf(s, "%2.02f KiB", (double)b / 1024);
	} else {
		sprintf(s, "%ld   B", b);
	}
}

void
screen_init(void)
{

	mainwindow = initscr();
	noecho();
	cbreak();
	nodelay(mainwindow, TRUE);
	refresh();
	wrefresh(mainwindow);
	screen = newwin(0, 0, 0, 0);
}

static void
update_display(int delay)
{
	char suptime[100];
	int u, days, hours, mins, secs;
	char stxtotal[100], srxtotal[100], stxnow[100], srxnow[100];
	char sdelay[100];

	u = uptime;
	days = u / 86400; u %= 86400;
	hours = u / 3600; u %= 3600;
	mins = u / 60; u %= 60;
	secs = u;

	if (uptime >= 86400) {
		sprintf(suptime,
		    "%d day%s, %d hour%s %d minute%s and %d second%s",
		    days, days == 1 ? "" : "s",
		    hours, hours == 1 ? "" : "s",
		    mins, mins == 1 ? "" : "s",
		    secs, secs == 1 ? "" : "s");
	} else if (uptime >= 3600) {
		sprintf(suptime,
		    "%d hour%s %d minute%s and %d second%s",
		    hours, hours == 1 ? "" : "s",
		    mins, mins == 1 ? "" : "s",
		    secs, secs == 1 ? "" : "s");
	} else if (uptime >= 60) {
		sprintf(suptime,
		    "%d minute%s and %d second%s",
		    mins, mins == 1 ? "" : "s",
		    secs, secs == 1 ? "" : "s");
	} else {
		sprintf(suptime,
		    "%d second%s",
		    secs, secs == 1 ? "" : "s");
	};
	sdelay[0] = 0;
	if (delay > 5) {
		sprintf(sdelay, "(no data for %d seconds)", delay);
	}

	bytestosomething(rxtotal, srxtotal);
	bytestosomething(txtotal, stxtotal);
	bytestosomething(rxnow, srxnow);
	bytestosomething(txnow, stxnow);

	int X = getmaxx(screen);
	int Y = getmaxy(screen);

	werase(screen);
	curs_set(0);
	mvwprintw(screen, 0, 0, "E169 stats (%d x %d) %s", X, Y, sdelay);
	mvwprintw(screen, 1, 0, "Time online: %s", suptime, delay);
	mvwprintw(screen, 2, 0, "Mode: %s / SRVST: %d", mode, srvst);
	if (rrssi == RSSI_NOTHING)
		mvwprintw(screen, 3, 0, "RSSI:  no data");
	else
		mvwprintw(screen, 3, 0, "RSSI:  %d dBm (%d)", crssi, rrssi);
	mvwprintw(screen, 4, 0, "Total: %s / %s", stxtotal, srxtotal);
	mvwprintw(screen, 5, 0, "Now:   %s / %s", stxnow, srxnow);

	for (i = -9; i <= 0; i++) {
		char buf[100], srx[100], stx[100];
		struct tm *tm;
		time_t now;

		now = time(NULL) + i * 60;
		tm = localtime(&now);
		strftime(buf, 100, "%H:%M", tm);

		bytestosomething(arx[ARX + i - 1], srx);
		bytestosomething(atx[ARX + i - 1], stx);

		mvwprintw(screen, 11 + i, 39, "%s %12s / %12s", buf, stx, srx);
	}

	for (i = ARSSI - X; i < ARSSI; i++) {
		#define DIVISOR 2
		static char c[3] = "_-~";
		int mod = arssi[i] % DIVISOR;

		if (arssi[i] == RSSI_NOTHING)
			continue;
		mvwprintw(screen,
		    Y - 1 - (arssi[i] / DIVISOR),
		    X + i - ARSSI,
		    "%c", c[mod]);
	}
	wrefresh(screen);
}

void
screen_end(void)
{

	endwin();
}


int
main(int argc, char **argv)
{
	FILE *fin, *fout;
	int fd;
	char line[100];
	int i;
	fd_set fdset;
	struct timeval timeout;
	int empty = 0;
	char *dev = "/dev/cuaU0.2";

	for (i = 0; i < ARSSI; i++) {
		arssi[i] = RSSI_NOTHING;
	}

	if ((fout = fopen("log", "w")) == NULL) {
		perror("opening fout");
		exit(0);
	}
	if (argc == 2) dev = argv[1];
	if ((fin = fopen(dev, "r")) == NULL) {
		perror(dev);
		exit(0);
	}
	fd = fileno(fin);

	screen_init();
	while (1) {
		update_display(empty);

		FD_ZERO(&fdset);
		FD_SET(fd, &fdset);
		timeout.tv_sec = 1;
		timeout.tv_usec = 0;
		if (select(fd + 1, &fdset, NULL, NULL, &timeout) < 0) {
			if (errno != EINTR) {
				perror("select");
				screen_end();
				exit(1);
			}
			continue;
		}
		if (!FD_ISSET(fd, &fdset)) {

			/* Shift anyway, the time should go on */
			memmove(arssi, arssi + 1, sizeof(int) * (ARSSI - 1));
			arssi[ARSSI - 1] = rrssi;

//			memmove(arx, arx + 1, sizeof(long) * (ARX - 1));
//			arx[ARX - 1] = 0;
//			memmove(atx, atx + 1, sizeof(long) * (ATX - 1));
//			atx[ATX - 1] = 0;

			empty++;
			if (empty > 5) {
				fprintf(fout, "Nothing for %d seconds!\n",
				    empty);
				fflush(fout);
			}
			continue;
		}
		empty = 0;
		FD_CLR(fd, &fdset);

		if (fgets(line, sizeof(line), fin) == NULL) {
			screen_end();
			perror("fgets");
			exit(1);
		}

		if (line[0] < ' ')
			continue;

		fprintf(fout, "%s", line);
		fflush(fout);

		if (strncmp(line, "^BOOT:", 6) == 0) 
			continue;

		if (strncmp(line, "^RSSI:", 6) == 0) {
			rrssi = atoi(line + 6);
			crssi = 2 * rrssi - 113;
			continue;
		}

		if (strncmp(line, "^SRVST:", 7) == 0) {
			srvst = atoi(line + 7);
			continue;
		}

		if (strncmp(line, "^DSFLOWRPT:", 11) == 0) {
			/* Time */
			line[19] = 0;
			uptime = strtol(line + 11, NULL, 16);

/*
0         1         2         3         4         5         6         7         8
012345678901234567890123456789012345678901234567890123456789012345678901234567890123456789
^DSFLOWRPT:0000071A,0000002A,0000002A,0000000000060BB1,00000000000CF3FB,00003E80,00003E80
*/
			/* TX Bytes in last two seconds */
			line[28] = 0;
			txnow = 0;
			if (line[20] == '0')
				txnow = strtol(line + 20, NULL, 16);
			/* RX Bytes in last two seconds */
			line[37] = 0;
			rxnow = 0;
			if (line[29] == '0')
				rxnow = strtol(line + 29, NULL, 16);
			/* TX Bytes in total */
			line[54] = 0;
			txtotal = strtol(line + 38, NULL, 16);
			/* RX Bytes in total */
			line[71] = 0;
			rxtotal = strtol(line + 55, NULL, 16);

			memmove(arssi, arssi + 1, sizeof(int) * (ARSSI - 1));
			arssi[ARSSI - 1] = rrssi;
			if (lastuptime + 60 < uptime) {
				int diffmins = (uptime - lastuptime) / 60;
				fprintf(fout, "%ld %ld %d\n", lastuptime,
				    uptime, diffmins);
				if (diffmins < 0)
					diffmins = 0;
				else if (diffmins >= min(ARX, ATX))
					diffmins = min(ARX, ATX) - 1;
				memmove(arx, arx + diffmins,
				    sizeof(int) * (ARX - diffmins));
				memmove(atx, atx + diffmins,
				    sizeof(int) * (ATX - diffmins));
				for (i = 1; i <= diffmins; i++) {
					arx[ARX - i] = 0;
					atx[ATX - i] = 0;
				}
				lastuptime = uptime;
			}
			arx[ARX - 1] += rxnow;
			atx[ATX - 1] += txnow;

			continue;
		}
		if (strncmp(line, "^MODE:", 6) == 0) {
			char smode[4];
			strncpy(smode, line + 6, 3);
			smode[3] = 0;
			strcpy(mode, smode);
			if (strcmp(smode, "0,0") == 0) strcpy(mode, "None");
			if (strcmp(smode, "3,2") == 0) strcpy(mode, "Green");
			if (strcmp(smode, "5,4") == 0) strcpy(mode, "Blue");
			if (strcmp(smode, "5,5") == 0) strcpy(mode, "Cyan");
			continue;
		}

		fprintf(fout, "No idea what that is\n");
	}
	screen_end();
	return 0;
}
